import numpy as np
import json
from pathlib import Path

try:
    from PythonTools.raw2py import raw2py
    from PythonTools.ezrt_header import EzrtHeader
    from PythonTools.rek2py import rek2py
except ModuleNotFoundError:
    raise ModuleNotFoundError("Install PythonTools from Fraunhofer EZRT.")

from ..loader import (
    BaseDataLoader,
    PyProjection,
    PyProjectionGeometry,
    PyRegionOfIntrest,
    PyVolume,
)
from scipy.spatial.transform import Rotation

from ..rq_json.json_load import RqJsonLoader


class RawLoader(BaseDataLoader):
    def __init__(self):
        super().__init__(".raw", ".raw", ".roi-json", ".rek")

    def load_projection_geometry(self, load_path: Path) -> PyProjectionGeometry:
        header = EzrtHeader.fromfile(load_path)

        focal_spot_mm = np.array(header.agv_source_position) * 1000.0
        detector_postion_mm = np.array(header.agv_detector_center_position) * 1000.0

        line = np.array(header.agv_detector_line_direction)
        column = np.array(header.agv_detector_col_direction)
        normal = np.cross(line, column)

        matrix = np.eye(3)
        matrix[:, 0] = line
        matrix[:, 1] = column
        matrix[:, 2] = normal

        rotation = Rotation.from_matrix(matrix)
        detector_orientation_quad = rotation.as_quat()

        frame_id = header.number_of_images

        projection_geometry = PyProjectionGeometry(
            focal_spot_mm,
            detector_postion_mm,
            detector_orientation_quad,
            frame_id,
            focal_spot_orientation_quad=np.array([0.0, 0.0, 0.0, 1.0]),
        )

        return projection_geometry

    def load_projection(
        self, load_path: Path, switch_order: bool = True
    ) -> PyProjection:
        header, image = raw2py(load_path, switch_order=switch_order)

        focal_spot_mm = np.array(header.agv_source_position) * 1000.0
        detector_postion_mm = np.array(header.agv_detector_center_position) * 1000.0

        line = np.array(header.agv_detector_line_direction)
        column = np.array(header.agv_detector_col_direction)
        normal = np.cross(line, column)

        matrix = np.eye(3)
        matrix[:, 0] = line
        matrix[:, 1] = column
        matrix[:, 2] = normal

        rotation = Rotation.from_matrix(matrix)
        detector_orientation_quad = rotation.as_quat()

        frame_id = header.number_of_images

        detector_heigth_mm = header.pixel_width_in_um / 1000.0 * image.shape[1]
        detector_width_mm = header.pixel_width_in_um / 1000.0 * image.shape[0]

        voltage_kv = header.voltage_in_kv
        current_ua = header.current_in_ua
        exposure_time_ms = header.exposure_time_in_ms

        projection = PyProjection(
            focal_spot_mm,
            detector_postion_mm,
            detector_orientation_quad,
            image,
            detector_heigth_mm,
            detector_width_mm,
            frame_id,
            np.array([0.0, 0.0, 0.0, 1.0]),
            voltage_kv,
            current_ua,
            exposure_time_ms,
        )

        return projection

    def load_json(self, load_path: Path) -> dict:
        with open(str(load_path), "r") as f:
            data_dict = json.load(f)
        return data_dict

    def load_region_of_intrest(self, load_path: Path) -> PyRegionOfIntrest:
        RqJsonLoader.load_region_of_intrest(self, load_path)

    def load_volume(self, load_path: Path) -> PyVolume:
        header, volume = rek2py(load_path)
        load_path_roi = (
            load_path.parent / f"{load_path.stem}{self.region_of_intrest_suffix}"
        )
        roi = self.load_region_of_intrest(load_path_roi)

        if volume.dtype == np.uint16:
            data_type = 0
        elif volume.dtype == np.uint8:
            data_type = 1
        else:
            raise ValueError("data type not implemented.")

        return PyVolume(volume, roi, data_type)
