from __future__ import annotations

import numpy as np

from numpy import ndarray
from .projection_geometry import PyProjectionGeometry

from rq_interfaces.msg import Projection, ProjectionGeometry
from sensor_msgs.msg import Image
import ros2_numpy
from scipy.spatial.transform import Rotation


class PyProjection(PyProjectionGeometry):
    """
    Represents a projection including image data and associated geometry.

    Attributes:
        image (np.ndarray): The projection image data.
        detector_heigth_mm (float): The height of the detector in millimeters.
        detector_width_mm (float): The width of the detector in millimeters.
        voltage_kv (float): The voltage in kilovolts.
        current_ua (float): The current in microamperes.
        exposure_time_ms (float): The exposure time in milliseconds.
    """

    def __init__(
        self,
        focal_spot_mm: ndarray,
        detector_postion_mm: ndarray,
        detector_orientation_quad: ndarray,
        image: np.ndarray,
        detector_heigth_mm: float,
        detector_width_mm: float,
        frame_id: str = "object",
        focal_spot_orientation_quad: np.ndarray = np.array([0.0, 0.0, 0, 1.0]),
        voltage_kv: float = 100.0,
        current_ua: float = 100.0,
        exposure_time_ms: float = 1000.0,
    ) -> None:
        """
        Initializes a PyProjection instance.

        Args:
            focal_spot_mm (ndarray): Position of the focal spot in millimeters.
            detector_postion_mm (ndarray): Position of the detector in millimeters.
            detector_orientation_quad (ndarray): Orientation of the detector as a quaternion.
            image (np.ndarray): The projection image data.
            detector_heigth_mm (float): The height of the detector in millimeters.
            detector_width_mm (float): The width of the detector in millimeters.
            frame_id (str): Frame ID for the projection geometry. Default is 'object'.
            focal_spot_orientation_quad (np.ndarray): Orientation of the focal spot as a quaternion. Default is [0., 0., 0, 1.].
            voltage_kv (float): The voltage in kilovolts. Default is 100.
            current_ua (float): The current in microamperes. Default is 100.
            exposure_time_ms (float): The exposure time in milliseconds. Default is 1000.
        """

        super().__init__(
            focal_spot_mm,
            detector_postion_mm,
            detector_orientation_quad,
            frame_id,
            focal_spot_orientation_quad,
        )
        self.image = image.astype(np.uint16)
        self.detector_heigth_mm = detector_heigth_mm
        self.detector_width_mm = detector_width_mm
        self.voltage_kv = voltage_kv
        self.current_ua = current_ua
        self.exposure_time_ms = exposure_time_ms

    @classmethod
    def dummy(cls):
        """
        Creates a dummy instance of PyProjection for testing.

        Returns:
            PyProjection: A dummy instance with default values.
        """

        return cls(
            np.array([0.0, 100.0, 0]),
            np.array([0.0, -100.0, 0]),
            np.array([1.0, 0.0, 0, 1.0]),
            np.random.randint(0, 65535, size=(10, 10), dtype=np.uint16),
            10.0,
            10.0,
        )

    @classmethod
    def from_look_at(
        cls,
        focal_spot_mm: ndarray,
        detector_postion_mm: ndarray,
        image_shape: tuple[int],
        detector_heigth_mm: float,
        detector_width_mm: float,
        frame_id: str = "object",
        voltage_kv: float = 100.0,
        current_ua: float = 100.0,
        exposure_time_ms: float = 1000.0,
        up_vector: np.ndarray = np.array([0.0, 0.0, 1.0]),
    ) -> "PyProjection":
        image = np.zeros(image_shape, np.uint16)

        vector = focal_spot_mm - detector_postion_mm
        vector = vector / np.linalg.norm(vector)
        up_vector = up_vector / np.linalg.norm(up_vector)

        detector_orientation = np.eye(3)
        detector_orientation[:, 2] = vector
        detector_orientation[:, 1] = np.cross(up_vector, vector)
        detector_orientation[:, 0] = np.cross(
            detector_orientation[:, 1], detector_orientation[:, 2]
        )

        source_orientation = np.eye(3)
        source_orientation[:, 2] = -vector
        source_orientation[:, 1] = np.cross(up_vector, -vector)
        source_orientation[:, 0] = np.cross(
            source_orientation[:, 1], source_orientation[:, 2]
        )

        return cls(
            focal_spot_mm,
            detector_postion_mm,
            Rotation.from_matrix(detector_orientation).as_quat(),
            image,
            detector_heigth_mm,
            detector_width_mm,
            frame_id,
            Rotation.from_matrix(source_orientation).as_quat(),
            voltage_kv,
            current_ua,
            exposure_time_ms,
        )

    def look_at(
        self,
        focal_spot_mm: ndarray,
        detector_postion_mm: ndarray,
        up_vector: np.ndarray = np.array([0.0, 0.0, 1.0]),
    ) -> PyProjection:
        vector = focal_spot_mm - detector_postion_mm
        vector = vector / np.linalg.norm(vector)
        up_vector = up_vector / np.linalg.norm(up_vector)

        z_rot = Rotation.from_euler("Z", -90, True)

        detector_orientation = np.eye(3)
        detector_orientation[:, 2] = vector
        detector_orientation[:, 1] = np.cross(up_vector, vector)
        detector_orientation[:, 0] = np.cross(
            detector_orientation[:, 1], detector_orientation[:, 2]
        )

        detector_orientation_quad = (
            Rotation.from_matrix(detector_orientation) * z_rot
        ).as_quat()

        source_orientation = np.eye(3)
        source_orientation[:, 2] = -vector
        source_orientation[:, 1] = np.cross(up_vector, -vector)
        source_orientation[:, 0] = np.cross(
            source_orientation[:, 1], source_orientation[:, 2]
        )
        source_orientation_quad = (
            Rotation.from_matrix(source_orientation) * z_rot
        ).as_quat()

        return PyProjection(
            focal_spot_mm,
            detector_postion_mm,
            detector_orientation_quad,
            self.image,
            self.detector_heigth_mm,
            self.detector_width_mm,
            self.frame_id,
            source_orientation_quad,
            self.voltage_kv,
            self.current_ua,
            self.exposure_time_ms,
        )

    def __str__(self) -> str:
        """
        Returns a string representation of the PyProjection instance.

        Returns:
            str: The string representation of the projection.
        """
        print_str = "---\n"
        print_str += "PyProjection\n"
        print_str += "--- Projection Geometry:\n"
        print_str += (
            f"Detector Posiion [mm]:         \t {self.detector_postion_mm.tolist()} \n"
        )
        print_str += (
            f"Focal Spot [mm]:               \t {self.focal_spot_mm.tolist()} \n"
        )
        print_str += f"Detector Orientation [quad]:   \t {self.detector_orientation_quad.tolist()} \n"
        print_str += f"Focal Spot Orientation [quad]: \t {self.focal_spot_orientation_quad.tolist()} \n"
        print_str += "---\n\n"
        return print_str

    @classmethod
    def from_message(cls, msg: Projection):
        """
        Creates an instance of PyProjection from a ROS message.

        Args:
            msg (Projection): The ROS message containing projection data.

        Returns:
            PyProjection: An instance initialized from the ROS message.
        """
        focal_spot_mm = np.array(
            [
                msg.projection_geometry.focal_spot_postion_mm.x,
                msg.projection_geometry.focal_spot_postion_mm.y,
                msg.projection_geometry.focal_spot_postion_mm.z,
            ]
        )

        detector_center_mm = np.array(
            [
                msg.projection_geometry.detector_postion_mm.x,
                msg.projection_geometry.detector_postion_mm.y,
                msg.projection_geometry.detector_postion_mm.z,
            ]
        )

        detector_orientation_quad = np.array(
            [
                msg.projection_geometry.detector_orientation_quad.x,
                msg.projection_geometry.detector_orientation_quad.y,
                msg.projection_geometry.detector_orientation_quad.z,
                msg.projection_geometry.detector_orientation_quad.w,
            ]
        )

        focal_spot_orientation = np.array(
            [
                msg.projection_geometry.focal_spot_orientation_quad.x,
                msg.projection_geometry.focal_spot_orientation_quad.y,
                msg.projection_geometry.focal_spot_orientation_quad.z,
                msg.projection_geometry.focal_spot_orientation_quad.w,
            ]
        )

        detector_heigth_mm = msg.detector_heigth_mm
        detector_width_mm = msg.detector_width_mm
        frame_id = msg.projection_geometry.header.frame_id

        image = ros2_numpy.numpify(msg.image)

        voltage_kv = msg.voltage_kv
        current_ua = msg.current_ua
        exposure_time_ms = msg.exposure_time_ms

        return cls(
            focal_spot_mm,
            detector_center_mm,
            detector_orientation_quad,
            image,
            detector_heigth_mm,
            detector_width_mm,
            frame_id,
            focal_spot_orientation,
            voltage_kv,
            current_ua,
            exposure_time_ms,
        )

    def as_message(self) -> Projection:
        """
        Converts the PyProjection instance to a ROS message.

        Returns:
            Projection: The ROS message representing the projection.
        """
        message = Projection()
        projection_geometry = ProjectionGeometry()

        self.focal_spot_mm = self.focal_spot_mm.reshape((3,))
        self.detector_postion_mm = self.detector_postion_mm.reshape((3,))
        self.detector_orientation_quad = self.detector_orientation_quad.reshape((4,))
        self.focal_spot_orientation_quad = self.focal_spot_orientation_quad.reshape(
            (4,)
        )

        projection_geometry.focal_spot_postion_mm.x = float(self.focal_spot_mm[0])
        projection_geometry.focal_spot_postion_mm.y = float(self.focal_spot_mm[1])
        projection_geometry.focal_spot_postion_mm.z = float(self.focal_spot_mm[2])

        projection_geometry.detector_postion_mm.x = float(self.detector_postion_mm[0])
        projection_geometry.detector_postion_mm.y = float(self.detector_postion_mm[1])
        projection_geometry.detector_postion_mm.z = float(self.detector_postion_mm[2])

        projection_geometry.detector_orientation_quad.x = float(
            self.detector_orientation_quad[0]
        )
        projection_geometry.detector_orientation_quad.y = float(
            self.detector_orientation_quad[1]
        )
        projection_geometry.detector_orientation_quad.z = float(
            self.detector_orientation_quad[2]
        )
        projection_geometry.detector_orientation_quad.w = float(
            self.detector_orientation_quad[3]
        )

        projection_geometry.focal_spot_orientation_quad.x = float(
            self.focal_spot_orientation_quad[0]
        )
        projection_geometry.focal_spot_orientation_quad.y = float(
            self.focal_spot_orientation_quad[1]
        )
        projection_geometry.focal_spot_orientation_quad.z = float(
            self.focal_spot_orientation_quad[2]
        )
        projection_geometry.focal_spot_orientation_quad.w = float(
            self.focal_spot_orientation_quad[3]
        )

        message.projection_geometry = projection_geometry
        message.image = ros2_numpy.msgify(Image, self.image, "mono16")

        message.detector_heigth_mm = self.detector_heigth_mm
        message.detector_width_mm = self.detector_width_mm

        message.projection_geometry.header.frame_id = self.frame_id

        message.voltage_kv = self.voltage_kv
        message.current_ua = self.current_ua
        message.exposure_time_ms = self.exposure_time_ms

        return message

    def get_projection_geometry(self) -> PyProjectionGeometry:
        return PyProjectionGeometry(
            self.focal_spot_mm,
            self.detector_postion_mm,
            self.detector_orientation_quad,
            self.frame_id,
            self.focal_spot_orientation_quad,
        )

    @property
    def detector_heigth_px(self) -> int:
        return self.image.shape[0]

    @property
    def detector_width_px(self) -> int:
        return self.image.shape[1]

    @property
    def pixel_pitch_x_mm(self) -> float:
        return self.detector_width_mm / self.detector_width_px

    @property
    def pixel_pitch_y_mm(self) -> float:
        return self.detector_heigth_mm / self.detector_heigth_px

    @property
    def to_cera_transformation_matrix(self) -> np.ndarray:
        transformation_matrix = np.eye(3)
        # [mm] -> [px]
        transformation_matrix[2, 2] = 1.0 / self.pixel_pitch_x_mm
        transformation_matrix[0, 2] = -self.detector_width_px / 2.0
        transformation_matrix[1, 2] = -self.detector_heigth_px / 2.0

        return transformation_matrix

    @property
    def projection_matrix_cera_px(self):
        projection_geometry = self.get_projection_geometry()
        return (
            np.linalg.inv(self.to_cera_transformation_matrix)
            @ projection_geometry.projection_matrix
        )
