from numpy import ndarray
from numpy.core.multiarray import array as array
from rq_interfaces.msg import Volume
from visualization_msgs.msg import Marker
from .region_of_intrest import PyRegionOfIntrest
from sensor_msgs.msg import Image
from enum import IntEnum

import numpy as np
import ros2_numpy


class VOLUME_TYPES(IntEnum):
    """
    Enum representing volume data types.
    """

    UINT_16 = 0
    UINT_8 = 1


class PyVolume:
    """
    Represents a volumetric dataset with an associated region of interest.

    Attributes:
        roi (PyRegionOfIntrest): The region of interest associated with the volume.
        array (ndarray): The volumetric data array.
        data_typ (VOLUME_TYPES): The data type of the volume.
    """

    def __init__(
        self, np_array: ndarray, roi: PyRegionOfIntrest, data_type: VOLUME_TYPES = ...
    ):
        """
        Initializes a PyVolume instance.

        Args:
            array (ndarray): The volumetric data array.
            roi (PyRegionOfIntrest): The region of interest associated with the volume.
            data_type (VOLUME_TYPES): The data type of the volume. Default is VOLUME_TYPES.UINT_8.
        """

        self.roi = roi
        self.array = np_array
        self.data_typ = data_type

    @staticmethod
    def get_data_type(volume_type: VOLUME_TYPES) -> np.dtype:
        """
        Gets the numpy data type corresponding to the given volume type.

        Args:
            volume_type (VOLUME_TYPES): The volume data type.

        Returns:
            np.dtype: The corresponding numpy data type.
        """

        if volume_type == VOLUME_TYPES.UINT_16:
            return np.uint16
        elif volume_type == VOLUME_TYPES.UINT_8:
            return np.uint8
        else:
            raise ValueError("Datatype is not implemented")

    @staticmethod
    def enum_to_numpify(volume_type: VOLUME_TYPES) -> str:
        """
        Converts the volume type enum to a string for use in ROS2 numpy conversion.

        Args:
            volume_type (VOLUME_TYPES): The volume data type.

        Returns:
            str: The corresponding string for ROS2 numpy conversion.
        """

        if volume_type == VOLUME_TYPES.UINT_16:
            return "mono16"
        elif volume_type == VOLUME_TYPES.UINT_8:
            return "mono8"
        else:
            raise ValueError("Datatype is not implemented")

    @classmethod
    def dummy(cls) -> "PyVolume":
        """
        Creates a dummy instance of PyVolume for testing.

        Returns:
            PyVolume: A dummy instance with random data.
        """

        roi = PyRegionOfIntrest.dummy()
        np_array = np.random.randint(0, 255, size=roi.shape)
        data_type = VOLUME_TYPES.UINT_8
        return cls(np_array, roi, data_type)

    @classmethod
    def from_message(cls, msg: Volume) -> "PyVolume":
        """
        Creates an instance of PyVolume from a ROS message.

        Args:
            msg (Volume): The ROS message containing volume data.

        Returns:
            PyVolume: An instance initialized from the ROS message.
        """

        roi: Marker = msg.grid.region_of_intrest_stack.markers[0]
        center_points_mm = np.array(
            [roi.pose.position.x, roi.pose.position.y, roi.pose.position.z]
        )

        dimensions_mm = np.array([roi.scale.x, roi.scale.y, roi.scale.z])

        frame_id = roi.header.frame_id
        resolution_mm = np.array(
            [msg.grid.resolution.x, msg.grid.resolution.y, msg.grid.resolution.z]
        )

        py_roi = PyRegionOfIntrest(
            center_points_mm, dimensions_mm, frame_id, resolution_mm
        )

        data_typ = msg.datatype

        np_array = np.zeros(py_roi.shape, dtype=cls.get_data_type(data_typ))

        for i, slice in enumerate(msg.slices):
            np_array[:, :, i] = (
                ros2_numpy.numpify(slice)
                .reshape((py_roi.shape[0], py_roi.shape[1]))
                .astype(cls.get_data_type(data_typ))
            )

        return cls(np_array, py_roi, data_typ)

    def as_message(self) -> Volume:
        """
        Converts the PyVolume instance to a ROS message.

        Returns:
            Volume: The ROS message representing the volume.
        """

        message = Volume()

        message.datatype = self.data_typ
        message.grid = self.roi.as_message()
        message.slices = list()

        for z in range(self.array.shape[2]):
            message.slices.append(
                ros2_numpy.msgify(
                    Image,
                    self.array[:, :, z].astype(self.get_data_type(self.data_typ)),
                    self.enum_to_numpify(self.data_typ),
                )
            )

        return message

    @property
    def shape(self):
        """
        Returns the shape of the volumetric data array.

        Returns:
            tuple: The shape of the volume.
        """

        return self.array.shape
