from __future__ import annotations

from rq_hardware.hardware_client import HardwareClient
from rq_ddetection.defect_detection_client import DefectDetectionClient
from rq_reconstruction.reconstruction_client import ReconstructionClient
from rq_trajectory.trajectory_optimization_client import TrajectoryOptimizationClient

from rq_controller.common import PyProjection, PyRegionOfIntrest


import rclpy


class WorkflowNode:
    def __init__(self):
        self.hardware_interface = HardwareClient()
        self.defect_detection_interface = DefectDetectionClient()
        self.trajectory = TrajectoryOptimizationClient()
        self.reconstruction = ReconstructionClient()

    def defect_detection(
        self, projections: list[PyProjection]
    ) -> list[PyRegionOfIntrest]:
        future = self.defect_detection_interface.multi_projection_defect_detection(
            projections
        )

        rclpy.spin_until_future_complete(self.defect_detection_interface, future)

        response = future.result()
        roi_list = self.defect_detection_interface.response_2_py(response)

        return roi_list

    def aquire_projection(self, projection: PyProjection):
        future = self.hardware_interface.aquire_projection(projection)

        rclpy.spin_until_future_complete(self.hardware_interface, future)

        response = future.result()
        projection = self.hardware_interface.projection_response_2_py(response)

        return projection

    def check_reachability(self, projection: PyProjection):
        future = self.hardware_interface.check_reachability(projection)

        rclpy.spin_until_future_complete(self.hardware_interface, future)

        response = future.result()
        reached, cost, _ = self.hardware_interface.reachability_response_2_py(response)
        return reached, cost

    def get_volume(self, projection_stack: list[PyProjection], roi: PyRegionOfIntrest):
        future = self.reconstruction.get_volume(projection_stack, roi)

        rclpy.spin_until_future_complete(self.reconstruction, future)

        response = future.result()
        return self.reconstruction.response_2_py(response)

    def trajectory_update(self, current_projection: PyProjection):
        future = self.trajectory.trajectory_update(current_projection)

        rclpy.spin_until_future_complete(self.trajectory, future)

        response = future.result()
        return self.trajectory.trajectory_update_response_2_py(response)

    def trajectory_initialize(self, roi: list[PyRegionOfIntrest]):
        future = self.trajectory.trajectory_initialize(roi)

        rclpy.spin_until_future_complete(self.trajectory, future)


def main():
    print(
        "This minimal example need the nodes described in the rq_workflow/launch/echo_launch.py script."
    )
    rclpy.init()
    workflow = WorkflowNode()

    print("Aquire some scout views...")
    projection_list = list()
    for i in range(5):
        print(f" - At projection {i}")
        scan_pose = PyProjection.dummy()
        scan_pose.focal_spot_mm += i
        projection_list.append(workflow.aquire_projection(scan_pose))

    print("Detect errors in scout views ...")
    roi = workflow.defect_detection(projection_list)

    print("Init traj opt ...")
    workflow.trajectory_initialize(roi)

    print("Get next scan pose from traj opt ...")
    next_scan_pose, finished = workflow.trajectory_update(projection_list[-1])

    print("Aquire scan pose ...")
    projection_list.append(workflow.aquire_projection(next_scan_pose))

    print("Resconstruct projections ...")
    workflow.get_volume(projection_list, roi[0])

    print("FINISHED !!!")


if __name__ == "__main__":
    main()
