import pytest
import numpy as np
from rq_controller.common import (
    PyProjection,
)  # Adjust this import according to your module's path
from rq_interfaces.msg import Projection
from sensor_msgs.msg import Image
import ros2_numpy


@pytest.fixture
def example_message():
    msg = Projection()
    msg.projection_geometry.focal_spot_postion_mm.x = 10.0
    msg.projection_geometry.focal_spot_postion_mm.y = 20.0
    msg.projection_geometry.focal_spot_postion_mm.z = 30.0
    msg.projection_geometry.detector_postion_mm.x = 40.0
    msg.projection_geometry.detector_postion_mm.y = 50.0
    msg.projection_geometry.detector_postion_mm.z = 60.0
    msg.projection_geometry.detector_orientation_quad.x = 0.0
    msg.projection_geometry.detector_orientation_quad.y = 0.0
    msg.projection_geometry.detector_orientation_quad.z = 0.0
    msg.projection_geometry.detector_orientation_quad.w = 1.0
    msg.projection_geometry.focal_spot_orientation_quad.x = 0.0
    msg.projection_geometry.focal_spot_orientation_quad.y = 0.0
    msg.projection_geometry.focal_spot_orientation_quad.z = 0.0
    msg.projection_geometry.focal_spot_orientation_quad.w = 1.0
    msg.projection_geometry.header.frame_id = "test_frame"

    image_array = np.zeros((10, 10), dtype=np.uint16)
    msg.image = ros2_numpy.msgify(Image, image_array, encoding="mono16")

    msg.detector_heigth_mm = 100.0
    msg.detector_width_mm = 200.0
    msg.voltage_kv = 120.0
    msg.current_ua = 150.0
    msg.exposure_time_ms = 500.0
    return msg


def test_initialization():
    focal_spot_mm = np.array([1.0, 2.0, 3.0])
    detector_postion_mm = np.array([4.0, 5.0, 6.0])
    detector_orientation_quad = np.array([0.0, 0.0, 0.0, 1.0])
    image = np.zeros((10, 10), dtype=np.uint16)
    detector_heigth_mm = 10.0
    detector_width_mm = 20.0
    voltage_kv = 100.0
    current_ua = 200.0
    exposure_time_ms = 1000.0
    frame_id = "test_frame"
    focal_spot_orientation_quad = np.array([0.0, 0.0, 0.0, 1.0])

    projection = PyProjection(
        focal_spot_mm,
        detector_postion_mm,
        detector_orientation_quad,
        image,
        detector_heigth_mm,
        detector_width_mm,
        frame_id,
        focal_spot_orientation_quad,
        voltage_kv,
        current_ua,
        exposure_time_ms,
    )

    assert np.array_equal(projection.focal_spot_mm, focal_spot_mm)
    assert np.array_equal(projection.detector_postion_mm, detector_postion_mm)
    assert np.array_equal(
        projection.detector_orientation_quad, detector_orientation_quad
    )
    assert np.array_equal(
        projection.focal_spot_orientation_quad, focal_spot_orientation_quad
    )
    assert np.array_equal(projection.image, image)
    assert projection.detector_heigth_mm == detector_heigth_mm
    assert projection.detector_width_mm == detector_width_mm
    assert projection.voltage_kv == voltage_kv
    assert projection.current_ua == current_ua
    assert projection.exposure_time_ms == exposure_time_ms
    assert projection.frame_id == frame_id


def test_dummy_method():
    projection = PyProjection.dummy()

    assert np.array_equal(projection.focal_spot_mm, np.array([0.0, 100.0, 0.0]))
    assert np.array_equal(projection.detector_postion_mm, np.array([0.0, -100.0, 0.0]))
    assert np.array_equal(
        projection.detector_orientation_quad, np.array([1.0, 0.0, 0.0, 1.0])
    )
    assert projection.detector_heigth_mm == 10.0
    assert projection.detector_width_mm == 10.0
    assert projection.frame_id == "object"


def test_from_message(example_message):
    projection = PyProjection.from_message(example_message)

    assert np.array_equal(projection.focal_spot_mm, np.array([10.0, 20.0, 30.0]))
    assert np.array_equal(projection.detector_postion_mm, np.array([40.0, 50.0, 60.0]))
    assert np.array_equal(
        projection.detector_orientation_quad, np.array([0.0, 0.0, 0.0, 1.0])
    )
    assert np.array_equal(
        projection.focal_spot_orientation_quad, np.array([0.0, 0.0, 0.0, 1.0])
    )

    assert projection.detector_heigth_mm == 100.0
    assert projection.detector_width_mm == 200.0
    assert projection.voltage_kv == 120.0
    assert projection.current_ua == 150.0
    assert projection.exposure_time_ms == 500.0
    assert projection.frame_id == "test_frame"


def test_as_message(example_message):
    projection = PyProjection.from_message(example_message)
    msg = projection.as_message()

    assert msg.projection_geometry.focal_spot_postion_mm.x == 10.0
    assert msg.projection_geometry.focal_spot_postion_mm.y == 20.0
    assert msg.projection_geometry.focal_spot_postion_mm.z == 30.0
    assert msg.projection_geometry.detector_postion_mm.x == 40.0
    assert msg.projection_geometry.detector_postion_mm.y == 50.0
    assert msg.projection_geometry.detector_postion_mm.z == 60.0
    assert msg.projection_geometry.detector_orientation_quad.x == 0.0
    assert msg.projection_geometry.detector_orientation_quad.y == 0.0
    assert msg.projection_geometry.detector_orientation_quad.z == 0.0
    assert msg.projection_geometry.detector_orientation_quad.w == 1.0
    assert msg.projection_geometry.focal_spot_orientation_quad.x == 0.0
    assert msg.projection_geometry.focal_spot_orientation_quad.y == 0.0
    assert msg.projection_geometry.focal_spot_orientation_quad.z == 0.0
    assert msg.projection_geometry.focal_spot_orientation_quad.w == 1.0
    assert msg.projection_geometry.header.frame_id == "test_frame"
    assert np.array_equal(
        ros2_numpy.numpify(msg.image), np.zeros((10, 10), dtype=np.uint16)
    )
    assert msg.detector_heigth_mm == 100.0
    assert msg.detector_width_mm == 200.0
    assert msg.voltage_kv == 120.0
    assert msg.current_ua == 150.0
    assert msg.exposure_time_ms == 500.0


def test_properties():
    focal_spot_mm = np.array([1.0, 2.0, 3.0])
    detector_postion_mm = np.array([4.0, 5.0, 6.0])
    detector_orientation_quad = np.array([0.0, 0.0, 0.0, 1.0])
    image = np.zeros((20, 30), dtype=np.uint16)
    detector_heigth_mm = 100.0
    detector_width_mm = 200.0
    voltage_kv = 100.0
    current_ua = 200.0
    exposure_time_ms = 1000.0
    frame_id = "test_frame"
    focal_spot_orientation_quad = np.array([0.0, 0.0, 0.0, 1.0])

    projection = PyProjection(
        focal_spot_mm,
        detector_postion_mm,
        detector_orientation_quad,
        image,
        detector_heigth_mm,
        detector_width_mm,
        frame_id,
        focal_spot_orientation_quad,
        voltage_kv,
        current_ua,
        exposure_time_ms,
    )

    assert projection.detector_heigth_px == 20
    assert projection.detector_width_px == 30
    assert projection.pixel_pitch_x_mm == 200.0 / 30
    assert projection.pixel_pitch_y_mm == 100.0 / 20
