import pytest
import numpy as np
from rq_controller.common import (
    PyVolume,
    PyRegionOfIntrest,
)  # Adjust this import according to your module's path
from rq_controller.common.volume import VOLUME_TYPES
from rq_interfaces.msg import Volume
from visualization_msgs.msg import Marker
from sensor_msgs.msg import Image
import ros2_numpy


@pytest.fixture
def example_message():
    msg = Volume()

    marker = Marker()
    marker.pose.position.x = 10.0
    marker.pose.position.y = 20.0
    marker.pose.position.z = 30.0
    marker.scale.x = 40.0
    marker.scale.y = 50.0
    marker.scale.z = 60.0
    marker.header.frame_id = "test_frame"

    msg.grid.region_of_intrest_stack.markers.append(marker)

    msg.grid.resolution.x = 1.0
    msg.grid.resolution.y = 1.0
    msg.grid.resolution.z = 1.0

    msg.datatype = VOLUME_TYPES.UINT_16

    shape = (40, 50, 60)
    slices = np.random.randint(0, 65535, size=shape, dtype=np.uint16)
    for i in range(shape[2]):
        image_msg = ros2_numpy.msgify(Image, slices[:, :, i], encoding="mono16")
        msg.slices.append(image_msg)

    return msg


def test_initialization():
    roi = PyRegionOfIntrest.dummy()
    array = np.random.randint(0, 255, size=roi.shape, dtype=np.uint8)
    volume = PyVolume(array, roi, VOLUME_TYPES.UINT_8)

    assert np.array_equal(volume.array, array)
    assert volume.roi == roi
    assert volume.data_typ == VOLUME_TYPES.UINT_8


def test_dummy_method():
    volume = PyVolume.dummy()

    assert volume.array.shape == volume.roi.shape
    assert volume.data_typ == VOLUME_TYPES.UINT_8


def test_from_message(example_message):
    volume = PyVolume.from_message(example_message)

    assert volume.array.shape == (40, 50, 60)
    assert volume.data_typ == VOLUME_TYPES.UINT_16
    assert volume.roi.frame_id == "test_frame"


def test_as_message(example_message):
    volume = PyVolume.from_message(example_message)
    msg = volume.as_message()

    assert msg.datatype == VOLUME_TYPES.UINT_16
    assert len(msg.slices) == 60

    for i, slice_msg in enumerate(msg.slices):
        slice_array = ros2_numpy.numpify(slice_msg)
        assert np.array_equal(slice_array, volume.array[:, :, i])


def test_get_data_type():
    assert PyVolume.get_data_type(VOLUME_TYPES.UINT_16) == np.uint16
    assert PyVolume.get_data_type(VOLUME_TYPES.UINT_8) == np.uint8
    with pytest.raises(ValueError):
        PyVolume.get_data_type(-1)


def test_enum_to_numpify():
    assert PyVolume.enum_to_numpify(VOLUME_TYPES.UINT_16) == "mono16"
    assert PyVolume.enum_to_numpify(VOLUME_TYPES.UINT_8) == "mono8"
    with pytest.raises(ValueError):
        PyVolume.enum_to_numpify(-1)


def test_shape():
    roi = PyRegionOfIntrest.dummy()
    array = np.random.randint(0, 255, size=roi.shape, dtype=np.uint8)
    volume = PyVolume(array, roi, VOLUME_TYPES.UINT_8)

    assert volume.shape == array.shape
